#  Library to make crating bots for Twitch chat easier.
#  Copyright (c) 2019 Maciej Marciniak
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import inspect
import typing

import twitchirc

from .utils import _await_sync

Bot = typing.TypeVar('Bot')


class Command:
    def __init__(self, chat_command: str, function: typing.Callable, parent: Bot,
                 forced_prefix: typing.Optional[str] = None, enable_local_bypass: bool = True,
                 matcher_function: typing.Optional[typing.Callable[[twitchirc.ChannelMessage,
                                                                    typing.Any], bool]] = None,
                 limit_to_channels: typing.Optional[typing.List[str]] = None,
                 available_in_whispers: bool = True):
        """
        Representation of a command.

        :param chat_command: Text the command responds to.
        :param function: Handler for the command.
        :param parent: Bot parent.
        :param forced_prefix: (optional) Forced prefix, if any.
        :param enable_local_bypass: (default True) Allow bypassing of permissions by local moderators?
        :param matcher_function: (optional) Custom function to test if this command should be executed.
        :param limit_to_channels: (optional) Limit executions of the command to channels.
        :param available_in_whispers: (default True) Should the command be available in whispers?
        """
        self.available_in_whispers = available_in_whispers
        self.matcher_function = matcher_function
        self.enable_local_bypass = enable_local_bypass
        self.ef_command = (forced_prefix + chat_command + ' ') if forced_prefix is not None else chat_command + ' '
        self.chat_command = chat_command
        self.function = function
        self.permissions_required = []
        self.forced_prefix = forced_prefix
        self.parent = parent
        self.limit_to_channels = limit_to_channels
        self.no_whispers_message = 'This command is not available in whispers'

    def __call__(self, message: twitchirc.ChannelMessage):
        return _await_sync(self.acall(message))

    async def acall(self, message: typing.Union[twitchirc.ChannelMessage, twitchirc.WhisperMessage]):
        if self.limit_to_channels is not None and message.channel not in self.limit_to_channels:
            return
        if isinstance(message, twitchirc.WhisperMessage) and self.available_in_whispers is False:
            return self.no_whispers_message

        if self.permissions_required:
            o = self.parent.check_permissions_from_command(message, self)
            if o:  # a non-empty list of missing permissions.
                return
        if inspect.iscoroutinefunction(self.function):
            return await self.function(message)
        else:
            return self.function(message)



#  Library to make crating bots for Twitch chat easier.
#  Copyright (c) 2019 Maciej Marciniak
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.
import select
import typing

import twitchirc


def create_msg(text, channel) -> twitchirc.ChannelMessage:
    """
    Create a ChannelMessage with the provided text and channel.

    :return: Newly created ChannelMessage object.
    """
    msg = twitchirc.ChannelMessage(text=text, channel=channel, user='OUTGOING')
    msg.outgoing = True
    return msg


def connect(oauth_token, use_ssl=True) -> twitchirc.Connection:
    """
    Connect to IRC.

    :param oauth_token: Authentication token. You can get one at https://twitchapps.com/tmi/
    The username is not needed for connecting.
    :param use_ssl: You can enable or disable encryption here.
    :return: created connection.
    """
    conn = twitchirc.Connection('irc.chat.twitch.tv', port=6697 if use_ssl else 6667, secure=use_ssl)
    conn.connect('don_t_need_to_know_this', oauth_token)
    return conn


def connect_anon(use_ssl=True) -> twitchirc.Connection:
    """
    Connect to IRC.

    :param use_ssl: You can enable or disable encryption here.
    :return: created connection.
    """
    conn = twitchirc.Connection('irc.chat.twitch.tv', port=6697 if use_ssl else 6667, secure=use_ssl)
    conn.connect('justinfan1234', 'junstinfan1234')
    return conn


def recv(conn: twitchirc.Connection) -> typing.List[twitchirc.Message]:
    msgs = []
    while 1:
        ready, _, _ = select.select([conn.socket], [], [], 0)
        ready = bool(ready)
        if ready:
            conn.receive()
        m = conn.process_messages()
        if not m:
            break
        print(m)
        msgs.extend(m)

    return msgs


print(f'twitchirc shell shortcuts loaded. There are shortcuts here for making experimenting easier.')
print(' - create_msg(text: str, channel: str) -> Message \n'
      '     Creates a ChannelMessage with the provided text and channel, useful for sending test messages.')
print(' - connect(oauth_token: str, use_ssl=True) -> Connection\n'
      '     Connect.')
print(' - connect_anon(use_ssl=True) -> Connection\n'
      '     Connect anonymously.')
print(' - recv(conn: Connection) -> List[Message]\n'
      '     Receive and process all the messages waiting.')
if __name__ == '__main__':
    try:
        import IPython

        IPython.start_ipython()
    except ImportError:
        import code

        code.interact(local=globals())



#  Library to make crating bots for Twitch chat easier.
#  Copyright (c) 2019 Maciej Marciniak
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import copy
import json
import os

import twitchirc


class AmbiguousSaveError(Exception):
    def __init__(self, msg):
        self.msg = msg

    def __repr__(self):
        return f'AmbiguousSaveError({self.msg!r})'

    def __str__(self):
        return f'AmbiguousSaveError: {self.msg}'


class CannotLoadError(Exception):
    def __init__(self, msg):
        self.msg = msg

    def __repr__(self):
        return f'CannotLoadError({self.msg!r})'

    def __str__(self):
        return f'CannotLoadError: {self.msg}'


class Storage:
    def __init__(self, auto_save=False):
        self.data = {}
        self.auto_save = auto_save

    def _load(self):
        return {}

    def _save(self):
        pass

    def load(self):
        self.data = self._load()
        return self.data

    def save(self, is_auto_save=False):
        if is_auto_save and not self.auto_save:
            return
        self._save()

    def __repr__(self):
        return f'twitchirc.Storage(auto_save={self.auto_save})'

    def __str__(self):
        return '<Storage>'

    def __getitem__(self, item):
        return self.data.__getitem__(item)

    def __setitem__(self, key, value):
        self.data.__setitem__(key, value)
        self.save(is_auto_save=True)


class JsonStorage(Storage):
    def __repr__(self):
        return f'twitchirc.JsonStorage(file={self.file}, auto_save={self.auto_save})'

    def __str__(self):
        return '<JsonStorage>'

    def __init__(self, file: str, auto_save=False, default=None):
        super().__init__(auto_save)
        if default is None:
            default = {}

        self.file = os.path.abspath(file)
        try:
            twitchirc.info(f'Reading file {self.file}')
            self.load()
        except (CannotLoadError, json.decoder.JSONDecodeError) as e:
            twitchirc.warn(f'Failed to load JsonStorage: {e}')
            self.data = default

    def _load(self):
        if not os.path.isfile(self.file):
            raise CannotLoadError(f'Target path {self.file!r} is not a file.')
        with open(self.file, 'r') as f:
            data = json.load(f)
        self._old_data = copy.deepcopy(data)
        return data

    def _save(self):
        twitchirc.log('debug', f'Saving file {self.file!r}')
        try:
            with open(self.file, 'r') as f:
                try:
                    file_data = json.load(f)
                except json.decoder.JSONDecodeError:
                    pass
                else:
                    if file_data != self._old_data:
                        raise AmbiguousSaveError('Data in file on disk has changed.')
        except FileNotFoundError:
            pass
        with open(self.file, 'w') as f:
            json.dump(self.data, f)
        self._old_data = copy.deepcopy(self.data)
        twitchirc.log('debug', f'Saved file {self.file!r}')



#  Library to make crating bots for Twitch chat easier.
#  Copyright (c) 2019 Maciej Marciniak
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import datetime
import os
import re
import time

quiet = False
log_file = None
log_file_name = ''
log_file_counter = 0
log_start_time = datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")
log_rotate_delay = 1 * 24 * 60 * 60
# 1 day
# * 24 hours
# * 60 minutes
# * 60 seconds
log_rotate_timestamp = time.time() + log_rotate_delay
DISPLAY_LOG_LEVELS = {
    'info': 'info',
    'warn': 'warn',
    'err': 'err',
    'debug': 'debug'
}
LOG_FORMAT = '[{time}] [{level}] {message}\n'


def remove_other_logs():
    n = log_file_name.replace(".", "\\.")
    pat = re.compile(rf'{n}\.[0-9]+')
    p = os.path.dirname(log_file_name)
    for i in os.listdir('.' if p == '' else p):
        m = re.match(pat, i)
        if m:
            os.unlink(os.path.join(p, i))
            continue


def enable_file_logging(filename: str, rotation_delay=1 * 24 * 60 * 60):
    """Enable logging to files."""
    global log_file, log_file_name, log_rotate_timestamp, log_rotate_delay
    log_file_name = filename
    remove_other_logs()
    log_file = open(filename, 'w')
    log_rotate_delay = rotation_delay
    log_rotate_timestamp = time.time() + log_rotate_delay
    write_log_message()


def write_log_message():
    """Write the logging started message to file."""
    log_file.write(f'/*\n')
    log_file.write(f' * {datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")}\n')
    log_file.write(f' * Log file {log_file_name!r} (will be moved to {f"{log_file_name}.{log_file_counter}"!r})\n')
    log_file.write(f' * Program started on {log_start_time}\n')
    log_file.write(' */\n')
    log_file.flush()


def rotate_logs():
    """Rotate logs."""
    global log_rotate_timestamp, log_file, log_file_counter
    log_rotate_timestamp = time.time() + log_rotate_delay
    if log_file is not None:
        log_file.close()
        new_name = log_file_name + '.' + str(log_file_counter)

        os.rename(log_file_name, new_name)
        log_file_counter += 1
        log_file = open(log_file_name, 'w')
        write_log_message()


def log(level: str, *message: str, sep: str = ' '):
    """Write a message to logs."""
    if log_rotate_timestamp <= time.time():
        rotate_logs()
    msg = LOG_FORMAT.format(time=datetime.datetime.now().strftime("%H:%M:%S"),
                            level=DISPLAY_LOG_LEVELS[level],
                            message=sep.join(message))
    if log_file:
        log_file.write(msg)
        log_file.flush()
    if not quiet:
        print(msg, end='', flush=True)


def info(*message: str):
    """Alias to log('info', *message)"""
    log('info', *message)


def warn(*message: str):
    """Alias to log('warn', *message)"""
    log('warn', *message)



#  Library to make crating bots for Twitch chat easier.
#  Copyright (c) 2019 Maciej Marciniak
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""
Names for permissions.

:py:const:`LOCAL_BYPASS_PERMISSION_TEMPLATE` Template for local bypass permissions.

:py:const:`GLOBAL_BYPASS_PERMISSION` Permission for bypassing every other

:py:const:`PERMISSION_COMMAND_PERM` Permission for the stock !perm command.

:py:const:`PERMISSION_COMMAND_PERM_ADD` Permission for adding permissions using !perm -a.

:py:const:`PERMISSION_COMMAND_PERM_REMOVE` Permission for adding permissions using !perm -r.

:py:const:`PERMISSION_COMMAND_PERM_LIST` Permission for adding permissions using !perm -l.

:py:const:`PERMISSION_COMMAND_QUIT` Permission for the stock !quit command.

:py:const:`PERMISSION_COMMAND_PART` Permission for the stock !part command.

:py:const:`PERMISSION_COMMAND_PART_OTHER` Permission for parting other channels using the !part command.

:py:const:`PERMISSION_COMMAND_JOIN` Permission for the stock !join command.

:py:const:`GROUP_PARENT` Template for creating permissions for inheritance.
"""
LOCAL_BYPASS_PERMISSION_TEMPLATE = 'twitchirc.bypass.permission.local.{}'
GLOBAL_BYPASS_PERMISSION = 'twitchirc.bypass.permission'

PERMISSION_COMMAND_PERM = 'twitchirc.perm'
PERMISSION_COMMAND_PERM_ADD = 'twitchirc.perm.add'
PERMISSION_COMMAND_PERM_REMOVE = 'twitchirc.perm.remove'
PERMISSION_COMMAND_PERM_LIST = 'twitchirc.perm.list'

PERMISSION_COMMAND_QUIT = 'twitchirc.quit'
PERMISSION_COMMAND_PART = 'twitchirc.part'
PERMISSION_COMMAND_PART_OTHER = 'twitchirc.part.other'
PERMISSION_COMMAND_JOIN = 'twitchirc.join'

GROUP_PARENT = 'parent.{}'



#  Library to make crating bots for Twitch chat easier.
#  Copyright (c) 2019 Maciej Marciniak
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.
import asyncio
import sched
import select
import typing
import traceback

import twitchirc  # Import self.

from .connection import RECONNECT
from .utils import _await_sync


class Bot(twitchirc.Connection):
    def schedule_event(self, delay, priority, function, args: tuple, kwargs: dict):
        """
        Schedule an event using the scheduler. Wrapper for :py:meth:`shed.scheduler.enter`

        :param delay: Delay the task will be executed in.
        :param priority: Task's priority.
        :param function: Function that will be run.
        :param args: Arguments to that function.
        :param kwargs: Keyword arguments to that function.

        :return: Event object.
        """
        return self.scheduler.enter(delay, priority, function, args, kwargs)

    def schedule_event_absolute(self, time, priority, function, args: tuple, kwargs: dict):
        """
        Schedule an event using the scheduler with a time, not a delay. Wrapper for :py:meth:`shed.scheduler.enterabs`

        :param time: Time the task will be executed.
        :param priority: Task's priority.
        :param function: Function that will be run.
        :param args: Arguments to that function.
        :param kwargs: Keyword arguments to that function.

        :return: Event object.
        """
        return self.scheduler.enterabs(time, priority, function, args, kwargs)

    def schedule_repeated_event(self, delay, priority, function, args: tuple, kwargs: dict):
        """
        Schedule a repeated event using the scheduler.

        :param delay: Delay the task will be executed in.
        :param priority: Task's priority.
        :param function: Function that will be run.
        :param args: Arguments to that function.
        :param kwargs: Keyword arguments to that function.

        :return: Event object.

        NOTE: There is currently no way to cancel an event. You have to do it manually.
        """

        def run_event():
            function()
            self.scheduler.enter(delay, priority, run_event, args, kwargs)

        return self.scheduler.enter(delay, priority, run_event, args, kwargs)

    def run_commands_from_file(self, file_object):
        lines = file_object.readlines()
        user = 'rcfile'
        channel = 'rcfile'
        self._in_rc_mode = True
        for num, i in enumerate(lines):
            i: str = i.replace('\n', '')
            if i.startswith('@'):
                if i.startswith('@at'):
                    channel = i.replace('@at ', '')
                elif i.startswith('@as'):
                    user = i.replace('@as ', '')
                continue
            m = twitchirc.ChannelMessage(user=user, channel=channel, text=i)
            m.flags = {
                'badge-info': '',
                'badges': 'moderator/1',
                'display-name': 'RCFile',
                'id': '00000000-0000-0000-0000-{:0>12}'.format(num),
                'user-id': 'rcfile',
                'emotes': ''
            }
            self._call_command_handlers(m)
        self._in_rc_mode = False

    def __init__(self, address, username: str, password: typing.Union[str, None] = None, port: int = 6667,
                 no_connect=False, storage=None, no_atexit=False, secure=False, message_cooldown=3):
        """
        A bot class.

        :param address: Address to connect to
        :param username: Username to use
        :param password: Password if needed, otherwise None can be used.
        :param port: Irc port.
        :param no_connect: Don't connect to the chat straight away.
        :param no_atexit: Don't use atexit to automatically disconnect.
        :param storage: A :py:class:`twitchirc.Storage`.
        :param secure: Use a secure connection, SSL.
        :param message_cooldown: Per-channel cooldown for messages.
        """
        super().__init__(address, port, no_atexit=no_atexit, message_cooldown=message_cooldown, secure=secure)
        self.scheduler = sched.scheduler()
        self._in_rc_mode = False
        if not no_connect:
            self.connect(username, password)
        self.username = username
        self._password = password
        self.commands: typing.List[twitchirc.Command] = []
        self.handlers: typing.Dict[str, typing.List[typing.Callable]] = {
            'pre_disconnect': [],
            'post_disconnect': [],
            'pre_save': [],
            'post_save': [],
            'start': [],
            'recv_join_msg': [],
            'recv_part_msg': [],
            'recv_ping_msg': [],
            'permission_error': [],
            'any_msg': [],
            'chat_msg': []
        }
        """
        A dict of handlers

        Available handlers:

        +----------------+----------------------------------------+---------------------------------------------------+
        |name            |arguments                               |explanation                                        |
        +================+========================================+===================================================+
        |pre_disconnect  |()                                      |                                                   |
        +----------------+----------------------------------------+---------------------------------------------------+
        |post_disconnect |()                                      |                                                   |
        +----------------+----------------------------------------+---------------------------------------------------+
        |pre_save        |()                                      |                                                   |
        +----------------+----------------------------------------+---------------------------------------------------+
        |post_save       |()                                      |                                                   |
        +----------------+----------------------------------------+---------------------------------------------------+
        |start           |()                                      |                                                   |
        +----------------+----------------------------------------+---------------------------------------------------+
        |permission_error|(message, command, missing_permissions) | message -- the ChannelMessage during which this   |
        |                |                                        | permission_error was triggered.                   |
        |                |                                        |                                                   |
        |                |                                        | command -- the Command object that triggered it.  |
        |                |                                        |                                                   |
        |                |                                        | WARN: `command` can be None if `check_permissions`|
        |                |                                        | was called (not `check_permissions_from_command`).|
        |                |                                        |                                                   |
        |                |                                        | missing_permissions -- permissions that are       |
        |                |                                        | missing to run the command.                       |
        +----------------+----------------------------------------+---------------------------------------------------+
        |any_msg         |(message)                               |                                                   |
        +----------------+----------------------------------------+---------------------------------------------------+
        |chat_msg        |(message)                               |                                                   |
        +----------------+----------------------------------------+---------------------------------------------------+        
        """

        self.prefix = '!'
        """Bot prefix. Can be set to any string. Recommended to set before registering any commands"""
        self.storage = storage
        self.on_unknown_command = 'ignore'
        """
        Action to take when an unknown command is encountered.
        Warning: This doesn't apply to commands with a forced prefix.

        Available handlers:

        * ignore - ignore it (default)
        * warn - print a warning to stdout
        * chat_message - send a chat message saying that the command is invalid.
        """
        self.permissions = twitchirc.PermissionList()
        self.command_error_handler = None
        """
        A function to handle errors being raised.
        
        Called when a command fails.
        """
        self._tasks = []

    def add_command(self, command: str,
                    forced_prefix: typing.Optional[str] = None,
                    enable_local_bypass: bool = True,
                    required_permissions: typing.Optional[typing.List[str]] = None,
                    limit_to_channels: typing.Optional[typing.List[str]]=None,
                    available_in_whispers: bool = True) \
            -> typing.Callable[[typing.Callable[[twitchirc.ChannelMessage],
                                                typing.Any]], twitchirc.Command]:
        # I'm sorry if you are reading this definition
        # here's a better version
        #  -> ((twitchirc.ChannelMessage) -> Any) -> Command
        """
        Add a command to the bot.
        This function is a decorator.

        :param command: Command to be registered.
        :param forced_prefix: Force a prefix to on this command. This is useful when you can change the prefix in \
        the bot.
        :param enable_local_bypass: If False this function will ignore the permissions \
        `twitchirc.bypass.permission.local.*`. This is useful when creating a command that can change global settings.
        :param required_permissions: Permissions required to run this command.
        :param limit_to_channels: (optional) Limit executions of the command to channels.
        :param available_in_whispers: (default True) Should the command be available in whispers?

        This is a decorator.
        """

        if required_permissions is None:
            required_permissions = []

        def decorator(func: typing.Callable) -> twitchirc.Command:
            cmd = twitchirc.Command(chat_command=command,
                                    function=func, forced_prefix=forced_prefix, parent=self,
                                    enable_local_bypass=enable_local_bypass,
                                    limit_to_channels=limit_to_channels,
                                    available_in_whispers=available_in_whispers)
            cmd.permissions_required.extend(required_permissions)
            self.commands.append(cmd)
            self.call_middleware('add_command', (cmd,), cancelable=False)
            return cmd

        return decorator

    def check_permissions(self, message: twitchirc.ChannelMessage, permissions: typing.List[str],
                          enable_local_bypass=True):
        """
        Check if the user has the required permissions to run a command

        :param message: Message received.
        :param permissions: Permissions required.
        :param enable_local_bypass: If False this function will ignore the permissions \
        `twitchirc.bypass.permission.local.*`. This is useful when creating a command that can change global settings.
        :return: A list of missing permissions.

        NOTE `permission_error` handlers are called if this function would return a non-empty list.
        """
        o = self.call_middleware('permission_check', dict(user=message.user, permissions=permissions,
                                                          message=message), cancelable=True)
        if o is False:
            return ['impossible.event_canceled']

        if isinstance(o, tuple):
            return o[1]

        missing_permissions = []
        if message.user not in self.permissions:
            missing_permissions = permissions
        else:
            perms = self.permissions.get_permission_state(message)
            if twitchirc.GLOBAL_BYPASS_PERMISSION in perms or \
                    (enable_local_bypass
                     and twitchirc.LOCAL_BYPASS_PERMISSION_TEMPLATE.format(message.channel) in perms):
                return []
            for p in permissions:
                if p not in perms:
                    missing_permissions.append(p)
        if missing_permissions:
            self.call_handlers('permission_error', message, None, missing_permissions)
        return missing_permissions

    def check_permissions_from_command(self, message: twitchirc.ChannelMessage,
                                       command: twitchirc.Command):
        """
        Check if the user has the required permissions to run a command

        :param message: Message received.
        :param command: Command used.
        :return: A list of missing permissions.

        NOTE `permission_error` handlers are called if this function would return a non-empty list.
        """
        o = self.call_middleware('permission_check', dict(user=message.user, permissions=command.permissions_required,
                                                          message=message, command=command), cancelable=True)
        if o is False:
            return ['impossible.event_canceled']

        if isinstance(o, tuple):
            return o[1]

        missing_permissions = []
        if message.user not in self.permissions:
            missing_permissions = command.permissions_required
        else:
            perms = self.permissions.get_permission_state(message)
            if twitchirc.GLOBAL_BYPASS_PERMISSION in perms or \
                    (
                            command.enable_local_bypass
                            and (twitchirc.LOCAL_BYPASS_PERMISSION_TEMPLATE.format(message.channel) in perms)
                    ):
                return []
            for p in command.permissions_required:
                if p not in perms:
                    missing_permissions.append(p)
        if missing_permissions:
            self.call_handlers('permission_error', message, command, missing_permissions)
        return missing_permissions

    def _send_if_possible(self, message, source_message):
        if isinstance(message, str):
            self.send(source_message.reply(message))
        elif isinstance(message, twitchirc.ChannelMessage):
            self.send(message)
        elif isinstance(message, list):
            for item in message:
                self._send_if_possible(item, source_message)

    def _call_command_handlers(self, message: twitchirc.ChannelMessage):
        return _await_sync(self._acall_command_handlers(message))

    async def _call_command(self, handler, message):
        o = self.call_middleware('command', {'message': message, 'command': handler}, cancelable=True)
        if o is False:
            return
        t = asyncio.create_task(handler.acall(message))
        self._tasks.append({
            'task': t,
            'source_msg': message,
            'command': handler
        })
        await self._a_wait_for_tasks()

    async def _a_wait_for_tasks(self):
        if not self._tasks:
            return
        done, _ = await asyncio.wait({i['task'] for i in self._tasks}, timeout=0.1)
        # don't pause the bot for long unnecessarily.
        for task in done:
            t = None
            for elem in self._tasks:
                if elem['task'] == task:
                    t = elem
                    break
            if t is not None:
                try:
                    result = await t['task']
                except BaseException as e:

                    for line in traceback.format_exc(1000).split('\n'):
                        twitchirc.log('warn', line)

                    if self.command_error_handler is not None:
                        self.command_error_handler(e, t['command'], t['source_msg'])
                    self._tasks.remove(t)
                    continue
                self._send_if_possible(result, t['source_msg'])
                self._tasks.remove(t)

    async def _acall_command_handlers(self, message: typing.Union[twitchirc.ChannelMessage, twitchirc.WhisperMessage]):
        """Handle commands."""
        if message.text.startswith(self.prefix):
            was_handled = False
            if ' ' not in message.text:
                message.text += ' '
            for handler in self.commands:
                if callable(handler.matcher_function) and handler.matcher_function(message, handler):
                    await self._call_command(handler, message)
                    was_handled = True
                if message.text.startswith(self.prefix + handler.ef_command):
                    await self._call_command(handler, message)
                    was_handled = True
            if not was_handled:
                self._do_unknown_command(message)
        else:
            await self._acall_forced_prefix_commands(message)

    def _call_forced_prefix_commands(self, message):
        return _await_sync(self._acall_forced_prefix_commands(message))

    async def _acall_forced_prefix_commands(self, message):
        """Handle commands with forced prefixes."""
        for handler in self.commands:
            if handler.forced_prefix is None:
                continue
            elif message.text.startswith(handler.ef_command):
                await self._call_command(handler, message)

    def _do_unknown_command(self, message):
        """Handle unknown commands."""
        if self.on_unknown_command == 'warn':
            twitchirc.warn(f'Unknown command {message!r}')
        elif self.on_unknown_command == 'chat_message':
            msg = message.reply(f'Unknown command {message.text.split(" ", 1)[0]!r}')
            self.send(msg, msg.channel)
        elif self.on_unknown_command == 'ignore':
            # Just ignore it.
            pass
        else:
            raise Exception('Invalid handler in `on_unknown_command`. Valid options: warn, chat_message, '
                            'ignore.')

    async def arun(self):
        """
        This is a coroutine version of :py:meth:`run`.

        Connect to the server if not already connected. Process messages received.
        This function includes an interrupt handler that automatically calls :py:meth:`stop`.

        :return: nothing.
        """
        try:
            await self._arun()
        except KeyboardInterrupt:
            print('Got SIGINT, exiting.')
            self.stop()
            return

    def run(self):
        """
        Connect to the server if not already connected. Process messages received.
        This function includes an interrupt handler that automatically calls :py:meth:`stop`.

        :return: nothing.
        """
        try:
            self._run()
        except KeyboardInterrupt:
            print('Got SIGINT, exiting.')
            self.stop()
            return

    def _select_socket(self):
        """Check if the socket has any data to receive."""
        sel_output = select.select([self.socket], [], [], 0.1)
        return bool(sel_output[0])

    async def _run_once(self):
        """
        Do everything needed to run, but don't loop. This can be used as a non-blocking version of :py:meth:`run`.

        :return: False if the bot should quit, True if everything went right, RECONNECT if the bot needs to reconnect.
        """
        if self.socket is None:  # self.disconnect() was called.
            return False
        if not self._select_socket():  # no data in socket, assume all messages where handled last time and return
            return True
        twitchirc.log('debug', 'Receiving.')
        o = self.receive()
        if o == RECONNECT:
            return RECONNECT
        twitchirc.log('debug', 'Processing.')
        self.process_messages(100, mode=2)  # process all the messages.
        twitchirc.log('debug', 'Calling handlers.')
        for i in self.receive_queue.copy():
            twitchirc.log('debug', '<', repr(i))
            self.call_handlers('any_msg', i)
            if isinstance(i, twitchirc.PingMessage):
                self.force_send('PONG {}\r\n'.format(i.host))
                if i in self.receive_queue:
                    self.receive_queue.remove(i)
                continue
            elif isinstance(i, twitchirc.ReconnectMessage):
                self.receive_queue.remove(i)
                return RECONNECT
            elif isinstance(i, twitchirc.ChannelMessage):
                self.call_handlers('chat_msg', i)
                await self._acall_command_handlers(i)
            elif isinstance(i, twitchirc.WhisperMessage):
                await self._acall_command_handlers(i)
            elif isinstance(i, twitchirc.PongMessage):
                self._on_receive_pong(i)
            if i in self.receive_queue:  # this check may fail if self.part() was called.
                self.receive_queue.remove(i)
        if not self.channels_connected:  # if the bot left every channel, stop processing messages.
            return False
        self.flush_queue(max_messages=100)
        return True

    async def _arun(self):
        """
        Brains behind :py:meth:`run`. Doesn't include the `KeyboardInterrupt` handler.

        :return: nothing.
        """
        if self.socket is None:
            self.connect(self.username, self._password)
        self.hold_send = False
        self.call_handlers('start')
        while 1:
            run_result = await self._run_once()
            if run_result is False:
                twitchirc.log('debug', 'break')
                break
            if run_result == RECONNECT:
                self.call_middleware('reconnect', (), False)
                self.disconnect()
                self.connect(self.username, self._password)
            self.scheduler.run(blocking=False)
            await self._a_wait_for_tasks()
            await asyncio.sleep(0)

    def _run(self):
        """
        Runs :py:meth:`_arun`.

        :return: nothing.
        """
        loop = asyncio.get_event_loop()
        loop.run_until_complete(self._arun())

    def call_handlers(self, event, *args):
        """
        Call handlers for `event`

        :param event: The event that happened. See `handlers`
        :param args: Arguments to give to the handler.

        :return: nothing.
        """
        if event not in ['any_msg', 'chat_msg']:
            twitchirc.log('debug', f'Calling handlers for event {event!r} with args {args!r}')
        for h in self.handlers[event]:
            h(event, *args)

    def disconnect(self):
        """
        Disconnect from the server.

        :return: nothing.
        """
        self.call_handlers('pre_disconnect')
        super().disconnect()
        self.call_handlers('post_disconnect')

    def stop(self):
        """
        Stop the bot and disconnect.
        This function force saves the `storage` and disconnects using :py:meth:`disconnect`

        :return: nothing.
        """
        if self.socket is None:  # Already disconnected.
            return
        self.call_handlers('pre_save')
        self.storage.save(is_auto_save=False)
        self.call_handlers('post_save')
        self.disconnect()

    def send(self, message: typing.Union[str, twitchirc.ChannelMessage], queue='misc') -> None:
        """
        Send a message to the server.

        :param message: message to send
        :param queue: Queue for the message to be in. This will be automatically overridden if the message is a \
        ChannelMessage. It will be set to :py:meth:`ChannelMessage.channel`.

        :return: nothing

        NOTE The message will not be sent instantly and this is intended. If you would send lots of messages Twitch
        will not forward any of them to the chat clients.
        """
        if self._in_rc_mode:
            if isinstance(message, twitchirc.ChannelMessage):
                twitchirc.log('info', f'[OUT/{message.channel}, {queue}] {message.text}')
            else:
                twitchirc.log('info', f'[OUT/?, {queue}] {message}')
        else:
            super().send(message, queue)



#  Library to make crating bots for Twitch chat easier.
#  Copyright (c) 2019 Maciej Marciniak
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

# PATTERN SNIPPETS
TWITCH_ARGS_PATTERN = r'(@[^ ]+)'
CHANNEL_PATTERN = r'([a-z0-9A-Z_]+)'
USER_PATTERN = r'([a-zA-Z0-9_-]+)!.+?'

# PATTERNS
PRIVMSG_PATTERN_TWITCH = fr'{TWITCH_ARGS_PATTERN} :{USER_PATTERN} PRIVMSG #{CHANNEL_PATTERN} :(.*)'
PING_MESSAGE_PATTERN = r'PING :(.*)'
PONG_MESSAGE_PATTERN = r'PONG :(.*)'
NOTICE_MESSAGE_PATTERN = fr'{TWITCH_ARGS_PATTERN} :(.+) NOTICE #{CHANNEL_PATTERN} :(.*)'
GLOBAL_NOTICE_MESSAGE_PATTERN = fr':(.+) NOTICE \* :(.*)'
JOIN_MESSAGE_PATTERN = rf':{USER_PATTERN} JOIN #{CHANNEL_PATTERN}'
PART_MESSAGE_PATTERN = rf':{USER_PATTERN} PART #{CHANNEL_PATTERN}'
USERNOTICE_MESSAGE_PATTERN = rf'{TWITCH_ARGS_PATTERN} :tmi.twitch.tv USERNOTICE #{CHANNEL_PATTERN}'
USERSTATE_MESSAGE_PATTERN = rf'{TWITCH_ARGS_PATTERN} :tmi.twitch.tv USERSTATE #{CHANNEL_PATTERN}'
WHISPER_MESSAGE_PATTERN = rf'{TWITCH_ARGS_PATTERN} :{USER_PATTERN} WHISPER {CHANNEL_PATTERN} :(.*)'
RECONNECT_MESSAGE_PATTERN = rf':tmi.twitch.tv RECONNECT'

# Delete the pattern snippets
del TWITCH_ARGS_PATTERN, CHANNEL_PATTERN, USER_PATTERN
__all__ = ['PRIVMSG_PATTERN_TWITCH', 'PING_MESSAGE_PATTERN', 'NOTICE_MESSAGE_PATTERN', 'JOIN_MESSAGE_PATTERN',
           'PART_MESSAGE_PATTERN', 'GLOBAL_NOTICE_MESSAGE_PATTERN', 'WHISPER_MESSAGE_PATTERN',
           'RECONNECT_MESSAGE_PATTERN', 'USERNOTICE_MESSAGE_PATTERN', 'USERSTATE_MESSAGE_PATTERN',
           'PONG_MESSAGE_PATTERN']



#  Library to make crating bots for Twitch chat easier.
#  Copyright (c) 2019 Maciej Marciniak
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.
import atexit
import socket
import ssl
import time
import typing

import twitchirc  # Import self.

RECONNECT = 'RECONNECT'


def to_bytes(obj, encoding='utf-8'):
    if isinstance(obj, str):
        return bytes(obj, encoding)
    else:
        return bytes(obj)


class Connection:
    def call_middleware(self, action, arguments, cancelable) -> typing.Union[bool, typing.Tuple[bool, typing.Any]]:
        """
        Call all middleware.

        :param action: Action to run.
        :param arguments: Arguments to give, depends on which action you use, for more info see AbstractMiddleware.
        :param cancelable: Can the event be canceled?
        :return: False if the event was canceled, True otherwise.
        """
        event = twitchirc.Event(action, arguments, source=self, cancelable=cancelable)
        canceler: typing.Optional[twitchirc.AbstractMiddleware] = None
        for m in self.middleware:
            m.on_action(event)
            if not canceler and event.canceled:
                canceler = m
        if event.canceled:
            twitchirc.log('debug', f'Event {action!r} was canceled by {canceler.__class__.__name__}.')
            return False
        if event.result is not None:
            return True, event.result
        return True

    def __init__(self, address: str, port: int = 6667, message_cooldown: int = 3, no_atexit=False,
                 secure: bool = False):
        """
        A bare-bones connection class.

        :param address: Address to connect to.
        :param port: Irc port.
        :param no_atexit: Don't use atexit to automatically disconnect.
        :param secure: Use a secure connection, SSL.
        :param message_cooldown: Per-channel cooldown for messages.
        """
        self.middleware: typing.List[twitchirc.AbstractMiddleware] = []
        """
        Enabled middleware. Each entry in this list should be an instance of a subclass of \
        :py:class:`twitchirc.AbstractMiddleware`
        """

        self.message_cooldown: int = message_cooldown
        """Per-channel cooldown for messages."""
        self.address: str = address
        """Address to connect to."""
        self.port: int = port
        self.socket: typing.Union[socket.socket, None] = None
        self.queue: typing.Dict[str, typing.List[bytes]] = {
            'misc': []
        }
        self.receive_queue: typing.List[
            typing.Union[twitchirc.Message, twitchirc.ChannelMessage, twitchirc.PingMessage]] = []
        self.receive_buffer: str = ''
        self.message_wait: typing.Dict[str, float] = {
            'misc': time.time()
        }
        self.hold_send = False
        """Stop sending messages for a bit, improper use of this may hang your program."""

        self.channels_connected = []
        """All channels that the Connection is listening to."""
        self.channels_to_remove = []
        self.secure = secure
        self.last_sent_messages: typing.Dict[str, str] = {
            # 'channel': 'message text.'
        }

        if not no_atexit:
            @atexit.register
            def close_self():
                try:
                    self.disconnect()
                    twitchirc.log('info', 'Automatic disconnect: ok.')
                except Exception as e:
                    twitchirc.log('info', f'Automatic disconnect: fail ({e})')

    def moderate(self, channel: str, user: typing.Optional[str] = None, message_id: typing.Optional[str] = None):
        """
        Construct a ModerationContainer targeting the channel, and optionally a user and message.

        :return: Newly created ModerationContainer
        """
        return twitchirc.ModerationContainer(message_id, user, channel, parent=self)

    def join(self, channel):
        """
        Join a channel.

        :param channel: Channel you want to join.
        :return: nothing.
        """
        channel = channel.lower().strip('#')
        o = self.call_middleware('join', dict(channel=channel), True)
        if o is False:
            return

        twitchirc.log('debug', 'Joining channel {}'.format(channel))
        self.force_send(f'JOIN #{channel}\r\n')
        self.queue[channel] = []
        self.message_wait[channel] = time.time()
        if channel not in self.channels_connected:
            self.channels_connected.append(channel)

    def part(self, channel):
        """
        Leave a channel

        :param channel: Channel you want to leave.
        :return: nothing.
        """
        channel = channel.lower().strip('#')

        o = self.call_middleware('part', dict(channel=channel), cancelable=True)
        if o is False:
            return

        twitchirc.log('debug', f'Departing from channel {channel}')
        self.force_send(f'PART #{channel}\r\n')

        self.channels_to_remove.append(channel)

        while channel in self.channels_connected:
            self.channels_connected.remove(channel)

    def disconnect(self):
        """
        Disconnect from IRC.

        :return: nothing.
        """
        self.call_middleware('disconnect', {}, cancelable=False)
        try:
            self.socket.send(b'quit\r\n')
            self.socket.shutdown(socket.SHUT_WR)
            self.socket.close()
        except BrokenPipeError:
            pass  # connection is already dead
        self.socket = None

    def twitch_mode(self):
        self.cap_reqs()

    def cap_reqs(self, use_membership=True):
        """
        Send CAP REQs.

        :param use_membership: Send the membership capability.
        :return: nothing.
        """
        twitchirc.log('debug', f'Sending CAP REQs. Membership: {use_membership}')
        self.force_send(f'CAP REQ :twitch.tv/commands twitch.tv/tags'
                        f'{" twitch.tv/membership" if use_membership else ""}\r\n')

    def connect(self, username, password: typing.Union[str, None] = None) -> None:
        """
        Connect to the IRC server.

        :param username: Username that will be used.
        :param password: Password to be sent. If None the PASS packet will not be sent.
        """
        self.call_middleware('connect', dict(username=username), cancelable=False)
        twitchirc.info('Connecting...')
        self._connect()
        twitchirc.log('debug', 'Logging in...')
        self._login(username, password)
        twitchirc.log('debug', 'OK.')

    def _login(self, username, password: typing.Union[str, None] = None):
        """
        Send NICK and PASS. For external use use :py:meth:`connect`.

        :return: nothing.
        """
        self.username = username
        self._password = password
        if password is not None:
            self.force_send('PASS {}\r\n'.format(password))
        else:
            self.force_send('PASS this_looks_like_a_password_ok?\r\n')
        self.force_send('NICK {}\r\n'.format(username))

    def _connect(self):
        """
        Connect to the IRC server. For external use use :py:meth:`connect`.

        :return: nothing.
        """
        if self.secure:
            sock = socket.socket()
            sock.connect((self.address, self.port))
            self._ssl_context = ssl.create_default_context()
            self.socket = self._ssl_context.wrap_socket(sock, server_hostname=self.address)
        else:
            self.socket = socket.socket()
            self.socket.connect((self.address, self.port))

    def send(self, message: typing.Union[str, twitchirc.ChannelMessage], queue='misc') -> None:
        """
        Queue a packet to be sent to the server.

        For sending a packet immediately use :py:meth:`force_send`.

        :param message: Message to be sent to the server.
        :param queue: Queue name. This will be automatically overridden if the message is a ChannelMessage. It will \
        be set to `message.channel`.
        :return: Nothing
        """
        o = self.call_middleware('send', dict(message=message, queue=queue), cancelable=True)
        if o is False:
            twitchirc.log('debug', str(message), ': canceled')
            return

        if isinstance(message, twitchirc.ChannelMessage):
            if message.user == 'rcfile':
                twitchirc.info(str(message))
                return
            queue = message.channel
            if message.channel in self.last_sent_messages and self.last_sent_messages[message.channel] == message.text:
                message.text += '\U000e0000'
            self.last_sent_messages[message.channel] = message.text

        if self.socket is not None or self.hold_send:
            twitchirc.log('debug', 'Queued message: {}'.format(message))
            if queue not in self.queue:
                self.queue[queue] = []
                self.message_wait[queue] = time.time()
            self.queue[queue].append(to_bytes(message, 'utf-8'))
        else:
            twitchirc.warn(f'Cannot queue message: {message!r}: Not connected.')

    def force_send(self, message: typing.Union[str, twitchirc.Message]):
        """
        Send a message immediately, without making it wait in the queue.
        For queueing a message use :py:meth:`send`.

        :param message: Message to be sent to the server.
        :return: Nothing
        """
        # Call it VIP queue if you wish. :)
        o = self.call_middleware('send', dict(message=message, queue='__force_send__'), cancelable=True)
        if o is False:
            return
        if isinstance(o, tuple):
            message = o[1]

        twitchirc.log('debug', 'Force send message: {!r}'.format(message))
        self.queue['misc'].insert(0, to_bytes(message, 'utf-8'))
        self.flush_single_queue('misc', no_cooldown=True)

    def _send(self, message: bytes):
        """Send some data straight to the socket. If it is none do nothing."""
        if self.socket is None:
            return
        self.socket.send(message)

    def flush_single_queue(self, queue, no_cooldown=False, max_messages=1, now=None):
        """
        Send waiting messages.

        :param queue: Queue to flush.
        :param no_cooldown: Don't wait for cooldowns.
        :param max_messages: Maximum messages to send.
        :param now: Current time.
        :return: Number of messages sent.
        """
        if now is None:
            now = time.time()
        if self.hold_send:
            return 0
        if self.message_wait[queue] > now and not no_cooldown:
            return 0
        sent = 0
        for num, message in enumerate(self.queue[queue][:max_messages]):
            twitchirc.info(f'Sending message {message!r}')
            self._send(message)
            sent += 1
            self.message_wait[queue] = now + self.message_cooldown
        self.queue[queue] = self.queue[queue][max_messages:]
        return sent

    def flush_queue(self, max_messages: int = 1) -> int:
        """
        Send waiting messages

        :param max_messages: Maximum amount messages to send.
        :return: Amount of messages sent.
        """
        if self.hold_send:
            return 0
        sent = 0
        now = time.time()
        for queue_name in self.queue:
            sent += self.flush_single_queue(queue_name, False, max_messages, now)
        return sent

    def receive(self):
        """Receive messages from the server and put them in the :py:attr:`receive_buffer`."""
        try:
            message = str(self.socket.recv(4096), 'utf-8', errors='ignore').replace('\r\n', '\n')
        except ConnectionResetError as e:
            twitchirc.log('warn', e)
            return RECONNECT
        if message == '':
            twitchirc.log('warn', 'Empty message')
            return RECONNECT
        self.receive_buffer += message

    def _remove_parted_channels(self):
        """
        Remove channels that have been parted when the respective queues got emptied.

        :return: nothing.
        """
        for i in self.channels_to_remove.copy():
            if not self.queue[i]:
                self.channels_to_remove.remove(i)
                del self.message_wait[i]
                del self.queue[i]

    def process_messages(self, max_messages: int = 1, mode=-1) -> typing.List[twitchirc.Message]:
        """
        Process the messages from self.receive_buffer
        Modes:
        * -1 every message. Nothing goes to self.receive_queue.
        * 0 chat messages. Other messages go to self.receive_queue.
        * 1 other messages. Chat messages go to self.receive_queue.
        * 2 do not return anything. Everything goes to self.receive_queue.

        :param max_messages: Maximum amount of messages to process.
        :param mode: What messages to return.
        :return: All messages specified by `mode`.
        """
        self._remove_parted_channels()
        messages_to_return = []
        messages_to_add_to_recv_queue = []
        for _ in range(max_messages):
            if '\n' not in self.receive_buffer:
                break
            message = self.receive_buffer.split('\n', 1)[0]
            self.receive_buffer = self.receive_buffer.replace(message + '\n', '', 1)
            # Remove `message` from the buffer.
            if message == '':
                continue
            m = twitchirc.auto_message(message, parent=self)
            o = self.call_middleware('receive', dict(message=m), True)
            if o is False:
                continue

            if isinstance(m, twitchirc.ChannelMessage):
                if mode in [-1, 0]:
                    messages_to_return.append(m)
                else:
                    messages_to_add_to_recv_queue.append(m)
            else:
                if mode in [-1, 1]:
                    messages_to_return.append(m)
                else:
                    messages_to_add_to_recv_queue.append(m)
        self.receive_queue.extend(messages_to_add_to_recv_queue)
        return messages_to_return

    def clone(self):
        """Creates a new connection with the same auth"""
        new = Connection(self.address, self.port, self.message_cooldown, secure=self.secure)
        new.connect(self.username, self._password)
        return new

    def clone_and_send_batch(self, message_batch: typing.List[typing.Union[str, twitchirc.ChannelMessage]]):
        """Creates a new connection with the same auth and queues all the provided messages."""
        new = self.clone()
        for i in message_batch:
            new.send(i)
        return new



#  Library to make crating bots for Twitch chat easier.
#  Copyright (c) 2019 Maciej Marciniak
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import typing

import twitchirc


def get_no_permission_generator(bot: twitchirc.Bot):
    def permission_error_handler(event, msg: twitchirc.ChannelMessage,
                                 command: twitchirc.Command,
                                 missing_permissions: typing.List[str]):
        del event
        if command is None:
            return
        bot.send(msg.reply(f"@{msg.user} You are missing permissions ({', '.join(missing_permissions)}) to use "
                           f"command {command.chat_command}."))

    bot.handlers['permission_error'].append(permission_error_handler)


def get_perm_command(bot: twitchirc.Bot):
    @bot.add_command(command='perm', forced_prefix=None, enable_local_bypass=True,
                     required_permissions=[twitchirc.PERMISSION_COMMAND_PERM])
    def command_perm(msg: twitchirc.ChannelMessage):
        p = twitchirc.ArgumentParser(prog='!perm', add_help=False)
        g = p.add_mutually_exclusive_group(required=True)
        g.add_argument('-a', '--add', metavar=('USER', 'PERMISSION'), nargs=2, dest='add')
        g.add_argument('-r', '--remove', metavar=('USER', 'PERMISSION'), nargs=2, dest='remove')
        g.add_argument('-l', '--list', metavar='USER', const=msg.user, default=None, nargs='?', dest='list')
        g.add_argument('-h', '--help', action='store_true', dest='help')
        args = p.parse_args(args=msg.text.split(' ')[1:])
        if args is None or args.help:
            usage = msg.reply(f'@{msg.user} {p.format_usage()}')
            bot.send(usage)
            return
        if args.add:
            if bot.check_permissions(msg, [twitchirc.PERMISSION_COMMAND_PERM_ADD], enable_local_bypass=False):
                bot.send(msg.reply(f"@{msg.user} You cannot use !perm -a, since you don't have"
                                   f"the {twitchirc.PERMISSION_COMMAND_PERM_ADD} permission"))
                return
            if args.add[0] not in bot.permissions:
                bot.permissions[args.add[0]] = []
            if args.add[1] not in bot.permissions[args.add[0]]:
                bot.permissions[args.add[0]].append(args.add[1])
                bot.send(msg.reply(f'@{msg.user} Given permission {args.add[1]} to user {args.add[0]}.'))
            else:
                bot.send(msg.reply(f'@{msg.user} User {args.add[0]} already has permission {args.add[1]}.'))
                return
        elif args.remove:
            if bot.check_permissions(msg, [twitchirc.PERMISSION_COMMAND_PERM_REMOVE], enable_local_bypass=False):
                bot.send(msg.reply(f"@{msg.user} You cannot use !perm -r, since you don't have"
                                   f"the {twitchirc.PERMISSION_COMMAND_PERM_REMOVE} permission"))
                return
            if args.remove[0] not in bot.permissions:
                bot.permissions[args.remove[0]] = []
            if args.remove[1] not in bot.permissions[args.remove[0]]:
                bot.send(msg.reply(f"@{msg.user} User {args.remove[0]} already "
                                   f"doesn't have permission {args.remove[1]}."))
                return
            else:
                bot.permissions[args.remove[0]].remove(args.remove[1])
                bot.send(msg.reply(f'@{msg.user} Removed permission {args.remove[1]} from user {args.remove[0]}.'))
                return
        elif args.list:
            if bot.check_permissions(msg, [twitchirc.PERMISSION_COMMAND_PERM_LIST]):
                bot.send(msg.reply(f"@{msg.user} You cannot use !perm -l, since you don't have"
                                   f"the {twitchirc.PERMISSION_COMMAND_PERM_LIST} permission"))
                return
            args.list = args.list.lower()
            if args.list not in bot.permissions:
                bot.permissions[args.list] = []
            if args.list == msg.user:
                output = ', '.join(bot.permissions.get_permission_state(msg))
                bot.send(msg.reply(f'You have permissions: {output}'))
            else:
                output = ', '.join(bot.permissions[args.list])
                bot.send(msg.reply(f'User {args.list} has permissions: {output}'))

            return


def get_quit_command(bot: twitchirc.Bot):
    @bot.add_command(command='quit', forced_prefix=None, enable_local_bypass=False,
                     required_permissions=[twitchirc.PERMISSION_COMMAND_QUIT])
    def command_quit(msg: twitchirc.ChannelMessage):
        bot.send(msg.reply('Quitting.'))
        bot.stop()


def get_part_command(bot: twitchirc.Bot):
    @bot.add_command(command='part', forced_prefix=None,
                     required_permissions=[twitchirc.PERMISSION_COMMAND_PART])
    def command_part(msg: twitchirc.ChannelMessage):
        p = twitchirc.ArgumentParser(prog='!part', add_help=False)
        p.add_argument('channel', metavar='CHANNEL', nargs='?', const=msg.channel, default=msg.channel)
        p.add_argument('-h', '--help', action='store_true', dest='help')
        args = p.parse_args(msg.text.split(' ')[1:])
        if args is None or args.help:
            usage = msg.reply(f'@{msg.user} {p.format_usage()}')
            bot.send(usage)
            return
        if args.channel == '':
            args.channel = msg.channel
        channel = args.channel.lower()
        if channel != msg.channel.lower() and bot.check_permissions(msg, [twitchirc.PERMISSION_COMMAND_PART_OTHER],
                                                                    enable_local_bypass=False):
            bot.send(msg.reply(f"Cannot part from channel {channel}: your permissions are not valid in that channel."))
            return
        if channel not in bot.channels_connected:
            bot.send(msg.reply(f'Not in {channel}'))
            return
        else:
            bot.send(msg.reply(f'Parting from {channel}'))
            m = twitchirc.ChannelMessage('Bye.', 'OUTGOING', channel)
            m.outgoing = True
            bot.send(m)
            bot.flush_single_queue(msg.channel)
            bot.part(channel)


def get_join_command(bot: twitchirc.Bot):
    @bot.add_command(command='join', forced_prefix=None, enable_local_bypass=False,
                     required_permissions=[twitchirc.PERMISSION_COMMAND_JOIN])
    def command_join(msg: twitchirc.ChannelMessage):
        chan = msg.text.split(' ')[1].lower()
        if chan in ['all']:
            bot.send(msg.reply(f'Cannot join #{chan}.'))
            return
        if chan in bot.channels_connected:
            bot.send(msg.reply(f'This bot is already in channel #{chan}.'))
        else:
            bot.send(msg.reply(f'Joining channel #{chan}.'))
            bot.join(chan)



#  Library to make crating bots for Twitch chat easier.
#  Copyright (c) 2019 Maciej Marciniak
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
# 
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
# 
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.
import asyncio


def _await_sync(awaitable):
    """Makes `await` work in synchronous functions. Make sure not to call from an async context."""
    loop = asyncio.get_event_loop()
    return loop.run_until_complete(awaitable)



#  Library to make crating bots for Twitch chat easier.
#  Copyright (c) 2019 Maciej Marciniak
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.
import collections
import typing

import twitchirc

ChannelSettings = collections.namedtuple('ChannelSettings', [
    'emote',
    'followers',
    'slow',
    'unique',
    'subscribers'
], defaults=(
    False,
    False,
    False,
    False,
    False
))
COMMANDS = {
    'emote': {
        True: '/emoteonly',
        False: '/emoteonlyoff',
        'default': None
    },
    'followers': {
        True: '/followers {}',
        False: '/followersoff',
        'default': ''
    },
    'slow': {
        True: '/slow {}',
        False: '/slowoff',
        'default': '5'
    },
    'unique': {
        True: '/r9kbeta',
        False: '/r9kbetaoff',
        'default': None
    },
    'subscribers': {
        True: '/subscribers',
        False: '/subscribersoff',
        'default': None
    }
}


class ModerationContainer:
    """
    Run moderation commands in the context of the user, channel and message
    """

    def __init__(self, target_message_id, target_user, target_channel, parent=None):
        self.parent: twitchirc.Connection = parent
        self.target_message_id = target_message_id
        self.target_user = target_user
        self.target_channel = target_channel

    @classmethod
    def from_message(cls, msg, parent=None):
        new = cls(msg.flags['id'], msg.user, msg.channel, parent)
        return new

    def delete(self):
        self.parent.send(self.format_delete())

    def format_delete(self):
        """
        Create a message with the /delete command in it.

        :return: ChannelMessage containing generated command.
        """
        if self.target_message_id is not None:
            return twitchirc.ChannelMessage(user='OUTGOING', text=f'/delete {self.target_message_id}',
                                            channel=self.target_channel, outgoing=True)
        else:
            raise RuntimeError("This ModerationContainer doesn't target a message.")

    def timeout(self, time: typing.Union[int, str], reason: typing.Optional[str] = None):
        self.parent.send(self.format_timeout(time, reason))

    def format_timeout(self, time: typing.Union[int, str], reason: typing.Optional[str] = None):
        """
        Create a message with the /timeout command in it.

        :param time: Time to block the user for.
        :param reason: Reason for this action.
        :return: ChannelMessage containing generated command.
        """
        if isinstance(time, str) and ((not time[-1].isalpha()) or time.isalpha()):
            raise ValueError('Time needs to be an int or string with the last character indicating the unit.')
        if self.target_user is not None:
            return twitchirc.ChannelMessage(user='OUTGOING', text=f'/timeout {self.target_user} {time}'
                                                                  f'{f" {reason}" if reason is not None else ""}',
                                            channel=self.target_channel, outgoing=True)
        else:
            raise RuntimeError("This ModerationContainer doesn't target a user.")

    def purge(self, reason: typing.Optional[str] = None):
        self.parent.send(self.format_purge(reason))

    def format_purge(self, reason: typing.Optional[str] = None):
        """
        Use `format_timeout` to create a message with a one second timeout.

        :param reason: Reason for this action.
        :return: ChannelMessage containing generated command.
        """
        if self.target_user is not None:
            return self.format_timeout(1, reason)
        else:
            raise RuntimeError("This ModerationContainer doesn't target a user.")

    def permaban(self, reason: typing.Optional[str] = None):
        self.parent.send(self.format_permaban(reason))

    def format_permaban(self, reason: typing.Optional[str] = None):
        """
        Create a message with the /ban command in it.

        :param reason: Reason for this action.
        :return: ChannelMessage containing generated command.
        """
        if self.target_user is not None:
            return twitchirc.ChannelMessage(user='OUTGOING', text=f'/ban {self.target_user}'
                                                                  f'{f" {reason}" if reason is not None else ""}',
                                            channel=self.target_channel, outgoing=True)
        else:
            raise RuntimeError("This ModerationContainer doesn't target a user.")

    def set_vip(self, status):
        self.parent.send(self.format_set_vip(status))

    def format_set_vip(self, status):
        """
        Create a message with the /[un]vip command in it.

        :param status: New status for the user. True for vip, False for unvip.
        :return: ChannelMessage containing generated command.
        """
        if self.target_user is not None:
            return twitchirc.ChannelMessage(user='OUTGOING',
                                            text=f'/{"un" if not status else ""}vip {self.target_user}',
                                            channel=self.target_channel, outgoing=True)
        else:
            raise RuntimeError("This ModerationContainer doesn't target a user.")

    def set_mod(self, status):
        self.parent.send(self.format_set_mod(status))

    def format_set_mod(self, status):
        """
        Run the [un]mod command on the targeted user.

        :param status: New status for the user. True for mod, False for unmod.
        :return: ChannelMessage containing generated command.
        """
        if self.target_user is not None:
            return twitchirc.ChannelMessage(user='OUTGOING',
                                            text=f'/{"un" if not status else ""}mod {self.target_user}',
                                            channel=self.target_channel, outgoing=True)
        else:
            raise RuntimeError("This ModerationContainer doesn't target a user.")

    def clear_channel(self):
        self.parent.send(self.format_clear_channel())

    def format_clear_channel(self):
        """
        Run the clear command on the targeted channel.

        :return: ChannelMessage containing generated command.
        """
        return twitchirc.ChannelMessage(user='OUTGOING', text=f'/clear',
                                        channel=self.target_channel, outgoing=True)

    def untimeout(self):
        self.parent.send(self.format_untimeout())

    def format_untimeout(self):
        """
        Create a message with the /untimeout command in it.

        :return: ChannelMessage containing generated command.
        """
        if self.target_user is not None:
            return twitchirc.ChannelMessage(user='OUTGOING', text=f'/untimeout {self.target_user}',
                                            channel=self.target_channel, outgoing=True)
        else:
            raise RuntimeError("This ModerationContainer doesn't target a user.")

    # noinspection PyArgumentList
    def set_channel_mode(self, settings: ChannelSettings,
                         old_settings=ChannelSettings()):
        for m in self.format_set_channel_mode(settings, old_settings):
            self.parent.send(m)

    # noinspection PyArgumentList
    def format_set_channel_mode(self, settings: ChannelSettings,
                                old_settings=ChannelSettings()) -> typing.List:
        """
        Generate commands needed to change

        :param settings: Settings to apply.
        :param old_settings: Settings from before. Used to not execute commands that are not needed.
        :return: A list of ChannelMessages containing commands.
        """
        commands = []
        for num, val in enumerate(settings):
            if val != old_settings[num]:
                c_entry = COMMANDS[settings._fields[num]]
                if isinstance(settings[num], bool):
                    replacement = ''
                else:
                    replacement = str(settings[num])
                text = (c_entry[settings[num] is not False]
                        .replace('{}',
                                 replacement))
                msg = twitchirc.ChannelMessage(text=text,
                                               user='OUTGOING', outgoing=True,
                                               channel=self.target_channel, parent=self.parent)
                commands.append(msg)
        return commands



#  Library to make crating bots for Twitch chat easier.
#  Copyright (c) 2019 Maciej Marciniak
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import typing

import twitchirc


def require_permission(permission: str):
    """
    Make the command require a permission.

    Warning: This is a decorator.

    :param permission: The permission
    """

    def decorator(func: twitchirc.Command) -> twitchirc.Command:
        if isinstance(func, twitchirc.Command):
            func.permissions_required.append(permission)
            return func
        else:
            raise Exception(f'Cannot call require_permission on object of type {type(func)!r}.')

    return decorator


def auto_group(message: twitchirc.ChannelMessage) -> str:
    """
    Get the user's group. This uses the Twitch badges.

    :param message: A message that was sent by the user.
    :return: group name
    """
    group = 'default'
    badges: typing.List[str] = message.flags['badges']
    if 'global_mod/1' in badges:
        return 'global_moderator'
    if 'staff/1' in badges:
        return 'staff'
    if 'broadcaster/1' in badges:
        return 'broadcaster'
    if 'moderator/1' in badges:
        return 'moderator'
    for i in badges:
        if i.startswith('subscriber'):
            group = 'subscriber'
    return group


class PermissionList:
    def __init__(self):
        """
        Class for storing permissions.

        Group hierarchy:

        .. code-block::

            DEFAULT
            |
            +-- moderator LOCAL_BYPASS
            |   |
            |   +-- global_moderator BYPASS
            |   |   |
            |   |   +-- admin BYPASS
            |   |       |
            |   |       +-- staff BYPASS
            |   |           |
            |   |           +-- bot_admin BYPASS
            |   |
            |   +-- broadcaster LOCAL_BYPASS
            |
            +-- subscriber
            |
            +-- vip

        """
        self.users = {

        }
        self.groups = {
            'default': [

            ],
            'moderator': [
                twitchirc.GROUP_PARENT.format('default')
                # 'parent.default'
            ],
            'subscriber': [
                twitchirc.GROUP_PARENT.format('default')
            ],
            'staff': [
                twitchirc.GROUP_PARENT.format('admin')
            ],
            'global_moderator': [
                twitchirc.GROUP_PARENT.format('moderator'),
                # 'parent.moderator',
                twitchirc.GLOBAL_BYPASS_PERMISSION
            ],
            'vip': [
                twitchirc.GROUP_PARENT.format('default')
                # 'parent.default'
            ],
            'broadcaster': [
                twitchirc.GROUP_PARENT.format('moderator')
                # 'parent.moderator'
            ],
            'admin': [
                twitchirc.GROUP_PARENT.format('global_moderator')
                # 'parent.global_moderator'
            ],
            'bot_admin': [
                twitchirc.GROUP_PARENT.format('staff')
                # 'parent.staff'
            ]
        }
        # DEFAULT
        # |
        # +-- moderator LOCAL_BYPASS
        # |   |
        # |   +-- global_moderator BYPASS
        # |   |   |
        # |   |   +-- admin BYPASS
        # |   |       |
        # |   |       +-- staff BYPASS
        # |   |           |
        # |   |           +-- bot_admin BYPASS
        # |   |
        # |   +-- broadcaster LOCAL_BYPASS
        # |
        # +-- subscriber
        # |
        # +-- vip

    def _get_permissions_from_parents(self, permissions: typing.List[str]):
        permissions = permissions.copy()
        while 1:
            was_extended = False
            for i in permissions.copy():
                if i.startswith('parent.'):
                    permissions.remove(i)
                    permissions.extend(self.groups[i.replace('parent.', '')])
                    was_extended = True
                    break
                if twitchirc.GLOBAL_BYPASS_PERMISSION in permissions:
                    return [twitchirc.GLOBAL_BYPASS_PERMISSION]
            if not was_extended:
                break
        return permissions

    def get_permission_state(self, message: twitchirc.ChannelMessage):
        """
        Get state of a user's permissions.

        :param message: Message that the user sent.
        :return: List of permissions the user has.
        """
        user = message.user
        group = auto_group(message)

        if user not in self.users:
            self.users[user] = []
        eff: typing.List[str] = self.groups[group].copy()

        eff: typing.List[str] = self._get_permissions_from_parents(eff)
        eff.extend(self.users[user])
        eff: typing.List[str] = self._get_permissions_from_parents(eff)
        if group in ['moderator', 'broadcaster']:
            eff.append(twitchirc.LOCAL_BYPASS_PERMISSION_TEMPLATE.format(message.channel))
            # eff.append(f'twitchirc.bypass.permission.local.{message.channel}')
        return eff

    def update(self, dict_: dict):
        """
        Update the permission list with entries from a dictionary. Keys have to be strings. Values have to be lists.
        If a key begins with `group.` it will be treated as a group, not a user.

        :return: nothing.
        """
        for k, v in dict_.items():
            if k.startswith('group.'):
                gn = k.replace('group.', '')
                if gn in self.groups:
                    self.groups[gn].extend(v)
                else:
                    self.groups[gn] = v
            else:
                if k in self.users:
                    self.users[k].extend(v)
                else:
                    self.users[k] = v

    def __iter__(self):
        for i in self.users:
            yield i
        for i in self.groups:
            yield f'group.{i}'

    def __dict__(self):
        output = {}
        output.update(self.users)
        for k, v in self.groups.items():
            output[f'group.{k}'] = v
        return output

    def __getitem__(self, item):
        if item.startswith('group.'):
            return self.groups[item.replace('group.', '')]
        else:
            return self.users[item]

    def __setitem__(self, key, value):
        if key.startswith('group.'):
            self.groups[key.replace('group.', '')] = value
        else:
            self.users[key] = value

    def fix(self):
        """Delete duplicate permissions."""
        for key in self:
            new = []
            for perm in self[key].copy():
                if perm in new:
                    continue
                new.append(perm)
            self[key] = new



#  Library to make crating bots for Twitch chat easier.
#  Copyright (c) 2019 Maciej Marciniak
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import argparse
import typing


class ArgumentParser(argparse.ArgumentParser):
    def __init__(self, *a, message_handler=lambda *a: None, **kw):
        super().__init__(*a, **kw)
        self.message_handler = message_handler

    def parse_args(self, args: typing.Optional[typing.Sequence[str]] = None,
                   namespace: typing.Optional[argparse.Namespace] = None) -> typing.Optional[argparse.Namespace]:
        try:
            return super().parse_args(args, namespace)
        except SystemExit:
            return

    def _print_message(self, message: str, file: typing.Optional[typing.IO[str]] = None) -> None:
        self.message_handler(message.rstrip('\n'))



#  Library to make crating bots for Twitch chat easier.
#  Copyright (c) 2019 Maciej Marciniak
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.
import typing


class Event:
    name: str
    data: typing.Dict[str, typing.Any]

    def __init__(self, name: str, data: typing.Dict[str, typing.Any], source: typing.Any, cancelable=True,
                 has_result=False):
        """
        Representation of an event.

        :param name: Event name
        :param data: Data provided with the event.
        :param source: Source of the event.
        :param cancelable: Can this event be canceled?
        :param has_result: Does this event have a replacable result?
        """
        self._cancelable = cancelable
        self._has_result = has_result
        self._lock = True

        self.name = name
        self.data = data
        self.canceled = False
        self.source = source

        self.result = None

    def __setattr__(self, key, value):
        if key in ['_cancelable', '_lock', '_has_result'] and hasattr(self, '_lock'):
            # make _lock lock the cancelable attribute.
            raise RuntimeError(f'Cannot assign to {key}')
        else:
            super().__setattr__(key, value)

    def cancel(self):
        """
        Cancel the event

        :return: nothing.
        :raises RuntimeError: If the event was canceled before.
        """
        if self.canceled:
            raise RuntimeError(f'Event {self.name} was canceled before.')
        self.canceled = True


# noinspection PyUnusedLocal,PyMethodMayBeStatic
class AbstractMiddleware:
    def __init__(self):
        """Base middleware class for Connections and Bots"""
        pass

    def __repr__(self):
        return f'<{self.__class__.__name__}()>'

    def __str__(self):
        return self.__repr__()

    def on_action(self, event: Event):
        """
        Called when an action is performed.

        :param event: Event
        :return: Action result
        """
        if event.name == 'send':
            self.send(event)
        elif event.name == 'receive':
            self.receive(event)
        elif event.name == 'command':
            self.command(event)
        elif event.name == 'permission_check':
            self.permission_check(event)
        elif event.name == 'join':
            self.join(event)
        elif event.name == 'part':
            self.part(event)
        elif event.name == 'disconnect':
            self.disconnect(event)
        elif event.name == 'connect':
            self.connect(event)
        elif event.name == 'add_command':
            self.add_command(event)
        elif event.name == 'reconnect':
            self.reconnect(event)
        # ignore unknown events.

    def send(self, event: Event) \
            -> None:
        """
        Called when sending a message.

        :param event: Event given. Required data keys: 'message'(str, bytes or a Message), optional: 'queue' (default \
        should be 'misc')
        Result: nothing.
        :return: Nothing
        """

    def receive(self, event: Event) -> None:
        """
        Called when receiving a message.

        :param event: Event given. Required data keys: 'message', optional: none. \
        Result: nothing.
        :return: Nothing
        """

    def command(self, event: Event) -> None:
        """
        Called when a command is run.

        :param event: Event given. Required data keys: 'message', 'command'. \
        Result: nothing.
        :return: Nothing
        """

    def permission_check(self, event: Event) -> None:
        """
        Called when permissions need to be checked.

        :param event: Event given. Required data keys: 'user', 'permissions', 'message', optional: 'command' \
        Result: Missing permissions

        :return: Nothing
        """

    def join(self, event: Event) -> None:
        """
        Called when joining a channel.

        :param event: Event given. Required data keys: 'channel'. \
        Result: nothing.
        :return: Nothing
        """

    def part(self, event: Event) -> None:
        """
        Called when parting a channel.

        :param event: Event given. Required data keys: 'channel'. \
        Result: nothing.
        :return: Nothing
        """

    def disconnect(self, event: Event) -> None:
        """
        Called when disconnecting from IRC.

        :param event: Event given. Required data keys: none. \
        Result: nothing.
        :return: Nothing
        """

    def connect(self, event: Event) -> None:
        """
        Called when connecting to IRC.

        :param event: Event given. Required data keys: 'username'. \
        Result: nothing.
        :return: Nothing
        """

    def add_command(self, event: Event) -> None:
        """
        Called when connecting to IRC.

        :param event: Event given. Required data keys: 'command'. \
        Result: nothing.
        :return: Nothing
        """

    def reconnect(self, event: Event) -> None:
        """
        Called when receiving a RECONNECT message from IRC or when getting kicked off.

        :param event: Event given. Required data keys: none, Result: nothing.
        :return: Nothing
        """



#  Library to make crating bots for Twitch chat easier.
#  Copyright (c) 2019 Maciej Marciniak
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.


# DO NOT BEAUTIFY THIS FILE
from .patterns import *
from .moderation import ModerationContainer
from .messages import (ChannelMessage, PartMessage, Message, JoinMessage, NoticeMessage, PingMessage, auto_message,
                       process_twitch_flags, PongMessage, WhisperMessage, ReconnectMessage, UserstateMessage,
                       UsernoticeMessage, GlobalNoticeMessage)
from .middleware import Event, AbstractMiddleware
from .connection import Connection
from .command import Command
from .bot import Bot

from .bot_storage import JsonStorage, CannotLoadError, AmbiguousSaveError
from .logging import (info, warn, log, log_file, log_rotate_delay, rotate_logs, enable_file_logging, LOG_FORMAT,
                      DISPLAY_LOG_LEVELS)
from .permissions import require_permission, auto_group, PermissionList
from .permission_names import *
from .stock_commands import (get_no_permission_generator, get_quit_command, get_part_command, get_join_command,
                             get_perm_command)
from .argument_parser import ArgumentParser



#  Library to make crating bots for Twitch chat easier.
#  Copyright (c) 2019 Maciej Marciniak
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import re
import typing

import twitchirc


def process_twitch_flags(flags) -> typing.Dict[str, typing.Union[str, typing.List[str]]]:
    """Process Twitch's tags."""
    # @<key>=<value>;[...]
    flags = flags[1:].split(';')
    # ['<key>=<value>', [...]]
    output = {}
    for i in flags:
        i = i.split('=', 1)
        val = i[1]
        if ',' in val:
            val = val.split(',')
        output[i[0]] = val
    return output


class Message:
    @staticmethod
    def from_match(m: typing.Match[str]):
        """
        Create a new object using a match

        :param m: Match
        :return: The new object
        """
        return Message(m.string)

    @staticmethod
    def from_text(text: str):
        """
        Create a new object from text

        :param text: Text to create it from.
        :return: The new object
        """
        new = Message('')
        new.raw_data = text
        if text.startswith(':'):
            new.source, new.action, new.args = text.split(' ', 2)
        else:
            new.args = text
        return new

    def __init__(self, args: str, outgoing=False, source=None, action='', parent=None, raw_data=None):
        """
        Message object.

        WARNING: If you receive this object at runtime, that means that, the packet you received is not known to this
        library

        :param args: Text received.
        """
        self.action = action
        self.source = source
        self.args: str = args
        self.outgoing = outgoing
        self.parent = parent
        self.raw_data = raw_data

    def __eq__(self, other):
        if isinstance(other, Message):
            return (other.args == self.args
                    and other.__class__ == self.__class__
                    and other.action == self.action
                    and other.source == self.source)
        else:
            return False

    def __repr__(self):
        if self.source is not None:
            return f'Message(source={self.source!r}, action={self.action!r}, args={self.args!r})'
        elif self.args:
            return f'Message(args={self.args!r})'
        elif self.raw_data:
            return f'Message(raw_data={self.raw_data!r})'
        else:
            return (f'Message({self.args!r}, source={self.source!r}, action={self.action!r}, '
                    f'raw_data={self.raw_data!r})')

    def __str__(self):
        return f'<Raw IRC message: {self.args!r}>'

    def __bytes__(self):
        if self.outgoing:
            if self.action:
                return bytes(self.action + ' ' + self.args)
            return bytes(self.args)
        else:
            return b''


class WhisperMessage(Message):
    @staticmethod
    def from_match(m: typing.Match[str]):
        flags = process_twitch_flags(m[1])
        new = WhisperMessage(flags=flags, user_from=m[2], user_to=m[3], text=m[4])
        new.raw_data = m[0]
        return new

    def __repr__(self):
        return (f'WhisperMessage(flags={self.flags!r}, user_from={self.user_from!r}, user_to={self.user_to!r}, '
                f'text={self.text!r})')

    def __str__(self):
        return f'{self.user_from} -> {self.user_to}: {self.text}'

    def __bytes__(self):
        if self.outgoing:
            return bytes(f'PRIVMSG #jtv :/w {self.user_to} {self.text}\r\n', 'utf-8')
        else:
            return b''

    def __init__(self, flags, user_from, user_to, text, outgoing=False):
        super().__init__(f'{"Outgoing" if outgoing else ""} WHISPER from {user_from} to {user_to}: {text}',
                         outgoing=outgoing)
        self.text = text
        self.user_to = user_to
        self.user_from = user_from
        self.flags = flags
        self.channel = 'whispers'

    @property
    def user(self):
        return self.user_from

    def reply(self, text: str):
        new = WhisperMessage(flags={}, user_from='OUTGOING', user_to=self.user_from, text=text, outgoing=True)
        return new


class ChannelMessage(Message):
    def moderate(self):
        if self.outgoing:
            raise RuntimeError('Cannot moderate a message that\'s going to be sent.')

        return twitchirc.ModerationContainer.from_message(self, self.parent)

    @staticmethod
    def from_match(m: typing.Match[str]):
        new = ChannelMessage(text=m[4], user=m[2], channel=m[3])
        new.flags = process_twitch_flags(m[1])
        new.raw_data = m[0]
        return new

    @staticmethod
    def from_text(text):
        m = re.match(twitchirc.PRIVMSG_PATTERN_TWITCH, text)
        if m:
            return ChannelMessage.from_match(m)
        else:
            raise ValueError(f'Cannot create a ChannelMessage from {text!r}')

    def __init__(self, text: str, user: str, channel: str, outgoing=False, parent=None):
        super().__init__(text, outgoing=outgoing, parent=parent)
        self._type = 'PRIVMSG'
        self.flags = {}
        self.text: str = text.replace('\r\n', '')
        self.user = user
        self.channel = channel

    def __repr__(self):
        return 'ChannelMessage(text={!r}, user={!r}, channel={!r})'.format(self.text, self.user, self.channel)

    def __str__(self):
        return '#{chan} <{user}> {text}'.format(user=self.user, text=self.text, chan=self.channel)

    def __bytes__(self):
        if self.outgoing:
            return bytes('PRIVMSG #{chan} :{text}\r\n'.format(chan=self.channel, text=self.text), 'utf-8')
        else:
            return b''

    def reply(self, text: str, force_slash=False):
        if not force_slash and text.startswith(('.', '/')):
            text = '/ ' + text
        new = ChannelMessage(text=text, user='OUTGOING', channel=self.channel)
        new.outgoing = True
        return new

    def reply_directly(self, text: str):
        new = WhisperMessage(flags={}, user_from='OUTGOING', user_to=self.user, text=text, outgoing=True)
        return new


class PingMessage(Message):
    @staticmethod
    def from_match(m: typing.Match[str]):
        new = PingMessage()
        new.host = m[1]
        new.raw_data = m[0]
        return new

    @staticmethod
    def from_text(text):
        m = re.match(twitchirc.PING_MESSAGE_PATTERN, text)
        return PingMessage.from_match(m)

    def __init__(self, host: typing.Optional[str] = None):
        self.host = host
        super().__init__(str(self))
        self.outgoing = False

    def __repr__(self):
        return 'PingMessage(host={!r})'.format(self.host)

    def __str__(self):
        return 'PING :{}'.format(self.host)

    def __bytes__(self):
        return bytes('PING :{}\r\n'.format(self.host), 'utf-8')

    def reply(self):
        return PongMessage(self.host)


class PongMessage(Message):
    @staticmethod
    def from_match(m: typing.Match[str]):
        return None

    @staticmethod
    def from_text(text):
        return None

    def __init__(self, host):
        super().__init__(host)
        self.host = host
        self.outgoing = True

    def __repr__(self):
        return 'PongMessage(host={!r})'.format(self.host)

    def __str__(self):
        return 'PONG :{}'.format(self.host)

    def __bytes__(self):
        return bytes('PONG :{}\r\n'.format(self.host), 'utf-8')

    def reply(self):
        raise RuntimeError('Cannot reply to a PongMessage.')


class NoticeMessage(Message):
    @staticmethod
    def from_match(m: typing.Match[str]):
        new = NoticeMessage('')
        new.message_id = process_twitch_flags(m[1])['msg-id']
        new.host = m[2]
        new.channel = m[3]
        new.text = m[4]
        # @msg-id=%s :tmi.twitch.tv NOTICE #{chan} :{msg}
        new.raw_data = m[0]
        return new

    @staticmethod
    def from_text(text):
        m = re.match(twitchirc.NOTICE_MESSAGE_PATTERN, text)
        if m:
            return NoticeMessage.from_match(m)
        else:
            raise ValueError('Invalid NoticeMessage(NOTICE #chan): {!r}'.format(text))

    def __init__(self, text, message_id=None, channel=None, host=None):
        super().__init__(text)
        self.text = text
        self.message_id = message_id
        self.channel = channel
        self.host = host


class GlobalNoticeMessage(NoticeMessage):
    @staticmethod
    def from_match(m: typing.Match[str]):
        new = GlobalNoticeMessage('')
        new.host = m[1]
        new.text = m[2]
        new.raw_data = m[0]
        return new

    @staticmethod
    def from_text(text):
        m = re.match(twitchirc.GLOBAL_NOTICE_MESSAGE_PATTERN, text)
        if m:
            return GlobalNoticeMessage.from_match(m)
        else:
            raise Exception('Invalid GlobalNoticeMessage(NOTICE *) {!r}'.format(text))

    def __init__(self, text, host=None):
        super().__init__(text, message_id=None, channel='*', host=host)


class JoinMessage(Message):
    @staticmethod
    def from_match(m: typing.Match[str]):
        new = JoinMessage(m[1], m[2])
        new.raw_data = m[0]
        return new

    def __init__(self, user: str, channel: str, outgoing=False):
        super().__init__('{} JOIN {}'.format(user, channel), outgoing=outgoing)
        self.user = user
        self.channel = channel

    def __repr__(self) -> str:
        if self.outgoing:
            return f'JoinMessage(user={self.user!r}, channel={self.channel!r}, outgoing=True)>'
        else:
            return f'JoinMessage(user={self.user!r}, channel={self.channel!r})>'

    def __str__(self):
        if self.outgoing:
            return f'JOIN {self.channel}'
        else:
            return f'{self.user} JOIN {self.channel}'

    def __bytes__(self):
        if self.outgoing:
            return bytes(f'JOIN {self.channel}\r\n', 'utf-8')
        else:
            return b''


class PartMessage(Message):
    @staticmethod
    def from_match(m: typing.Match[str]):
        new = PartMessage(m[1], m[2])
        new.raw_data = m[0]
        return new

    def __init__(self, user: str, channel: str, outgoing=False):
        super().__init__(f'{user} PART {channel}', outgoing=outgoing)
        self.user = user
        self.channel = channel

    def __repr__(self):
        if self.outgoing:
            return f'PartMessage(user={self.user!r}, channel={self.channel!r})  # outgoing'
        else:
            return f'PartMessage(user={self.user!r}, channel={self.channel!r})'

    def __str__(self):
        if self.outgoing:
            return f'<PART {self.channel}>'
        else:
            return f'<{self.user} PART {self.channel}>'

    def __bytes__(self):
        if self.outgoing:
            return bytes(f'PART #{self.channel}', 'utf-8')
        else:
            return b''


class UsernoticeMessage(Message):
    @staticmethod
    def from_match(m: typing.Match[str]):
        new = UsernoticeMessage(m[1], m[2])
        new.raw_data = m[0]
        return new

    def __repr__(self):
        return f'UsernoticeMessage(flags={self.flags!r}, channel={self.channel!r})'

    def __str__(self):
        return f'<USERNOTICE {self.channel}>'

    def __init__(self, flags, channel):
        super().__init__(flags + ' ' + channel)
        self.flags = process_twitch_flags(flags)
        self.channel = channel


class UserstateMessage(Message):
    @staticmethod
    def from_match(m: typing.Match[str]):
        new = UserstateMessage(m[1], m[2])
        new.raw_data = m[0]
        return new

    def __repr__(self):
        return f'UserstateMessage(flags={self.flags!r}, channel={self.channel!r})'

    def __str__(self):
        return f'<USERSTATE {self.channel}>'

    def __init__(self, flags, channel):
        super().__init__(flags + ' ' + channel)
        self.flags = process_twitch_flags(flags)
        self.channel = channel


class ReconnectMessage(Message):
    @staticmethod
    def from_match(m: typing.Match[str]):
        new = ReconnectMessage()
        new.raw_data = m[0]
        return new

    def __repr__(self):
        return f'ReconnectMessage()'

    def __str__(self):
        return f'<RECONNECT>'

    def __init__(self):
        super().__init__('RECONNECT')


MESSAGE_PATTERN_DICT: typing.Dict[typing.Any, typing.Union[
    typing.Type[ChannelMessage],
    typing.Type[PingMessage],
    typing.Type[NoticeMessage],
    typing.Type[GlobalNoticeMessage],
    typing.Type[JoinMessage],
    typing.Type[PartMessage],
    typing.Type[WhisperMessage],
    typing.Type[ReconnectMessage],
    typing.Type[UsernoticeMessage]
]
] = {
    re.compile(twitchirc.PRIVMSG_PATTERN_TWITCH): ChannelMessage,
    re.compile(twitchirc.PING_MESSAGE_PATTERN): PingMessage,
    re.compile(twitchirc.NOTICE_MESSAGE_PATTERN): NoticeMessage,
    re.compile(twitchirc.GLOBAL_NOTICE_MESSAGE_PATTERN): GlobalNoticeMessage,
    re.compile(twitchirc.JOIN_MESSAGE_PATTERN): JoinMessage,
    re.compile(twitchirc.PART_MESSAGE_PATTERN): PartMessage,
    re.compile(twitchirc.WHISPER_MESSAGE_PATTERN): WhisperMessage,
    re.compile(twitchirc.RECONNECT_MESSAGE_PATTERN): ReconnectMessage,
    re.compile(twitchirc.USERNOTICE_MESSAGE_PATTERN): UsernoticeMessage,
    re.compile(twitchirc.USERSTATE_MESSAGE_PATTERN): UserstateMessage,
    re.compile(twitchirc.PONG_MESSAGE_PATTERN): PongMessage
}


def auto_message(message, parent=None):
    for k, v in MESSAGE_PATTERN_DICT.items():
        m = re.match(k, message)
        if m:
            new = v.from_match(m)
            new.parent = parent
            return new

    # if nothing matches return generic irc message.
    return Message.from_text(message)



#  Library to make crating bots for Twitch chat easier.
#  Copyright (c) 2019 Maciej Marciniak
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

from setuptools import setup

setup(
    name='twitchirc',
    python_requires='>=3.7',
    version='1.7.8',
    packages=['twitchirc'],
    url='https://www.github.com/mm2pl/twitchirc',
    license='GPLv3',
    author='mm2pl',
    author_email='u.y@o2.pl',
    description='Unofficial library for Twitch irc.'
)
